"""Single-cell Embedded topic model.

An accurate, transferrable and interpretable model with highly scalable and
easy-to-use APIs for single-cell RNA sequencing data analysis. Fully integrated
with scanpy and anndata.

From paper "Learning interpretable cellular and gene signature embeddings from
single-cell transcriptomic data".
Link: https://www.biorxiv.org/content/10.1101/2021.01.13.426593v1.full

If you have any problems using this package, please refer to our project site
https://www.github.com/hui2000ji/scETM.
"""

from scETM.logging_utils import initialize_logger

initialize_logger()


import logging
import functools
import os

from numpy.lib.arraysetops import isin

logger = logging.getLogger(__name__)

def log_arguments(f):
    """Log the arguments of the function on calling.

    Args:
        f: the decorated function whose arguments will be logged.

    Returns:
        log_arguments_wrapper: a wrapped function that will log the arguments
            on calling.
    """

    if hasattr(f, '__qualname__'):
        name = f.__qualname__
    else:
        name = f.__name__
    @functools.wraps(f)
    def log_arguments_wrapper(*args, **kwargs):
        if name.endswith('__init__'):
            assert len(args) >= 1
            args_str = ', '.join(map(str, args[1:]))
        else:
            args_str = ', '.join(map(str, args))
        kwargs_str = ', '.join([f'{k} = {v}' for k, v in kwargs.items()])
        args_kwargs_str = ', '.join([s for s in [args_str, kwargs_str] if s])
        logger.info(f'{name}({args_kwargs_str})')
        result = f(*args, **kwargs)
        return result
    return log_arguments_wrapper

def initialize_logger(ckpt_dir=None, level=logging.INFO) -> None:
    """Initialize the scETM logger.

    Each time the function is called, close all file handlers of the scETM
    logger and create one logging to "ckpt_dir/log.txt" if ckpt_dir is not
    None.
    If the scETM logger has no handlers, create a formatted stream handler.

    Args:
        ckpt_dir: directory to store the log file. If None, no file handler
            will be added to the scETM logger.
        level: controls the level of logging.
    """

    logger_ = logging.getLogger('scETM')
    logger_.setLevel(level)
    if logger_.hasHandlers():
        for handler in logger_.handlers:
            if isinstance(handler, logging.FileHandler):
                logger_.warning(f'Reinitializing... The file handler {handler} will be closed.')
                logger_.removeHandler(handler)
    formatter = logging.Formatter('[%(asctime)s] %(levelname)s - %(name)s: %(message)s')
    if not logger_.hasHandlers():
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel(level)
        logger_.addHandler(stream_handler)
    if ckpt_dir is not None:
        file_handler = logging.FileHandler(os.path.join(ckpt_dir, 'log.txt'))
        file_handler.setFormatter(formatter)
        file_handler.setLevel(level)
        logger_.addHandler(file_handler)



from typing import Any, Mapping, Sequence, Tuple, Union, Iterable
import anndata
import numpy as np
import torch
from torch import nn, optim
from scETM.batch_sampler import CellSampler

class BaseCellModel(nn.Module):
    """Base class for single cell models.

    If you wish to modify scETM or implement other single cell models, consider
    extending this class.

    Attributes:
        clustering_input: name of the embedding used for clustering.
        emb_names: name of embeddings returned by self.forward.
        device: device to store the model parameters.
        n_trainable_genes: number of trainable genes.
        n_fixed_genes: number of fixed_genes. Parameters in the input and
            output layer related to these genes should be fixed. Useful for the
            fine-tuning stage in transfer learning.
        n_batches: number of batches in the dataset.
        need_batch: whether the model need batch infomation.
    """

    clustering_input: str
    emb_names: Sequence[str]

    def __init__(self,
        n_trainable_genes: int,
        n_batches: int,
        n_fixed_genes: int = 0,
        need_batch: bool = False,
        device: torch.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    ) -> None:
        """Initializes the BaseCellModel object.

        Args:
            n_trainable_genes: number of trainable genes.
            n_batches: number of batches in the dataset.
            n_fixed_genes: number of fixed_genes. Parameters in the input and
                output layer related to these genes should be fixed. Useful for the
                fine-tuning stage in transfer learning.
            need_batch: whether the model need batch infomation.
            device: device to store the model parameters.
        """

        super().__init__()
        self.device: torch.device = device
        self.n_trainable_genes: int = n_trainable_genes
        self.n_fixed_genes: int = n_fixed_genes
        self.n_batches: int = n_batches
        self.need_batch: bool = need_batch

    def train_step(self,
        optimizer: optim.Optimizer,
        data_dict: Mapping[str, torch.Tensor],
        hyper_param_dict: Mapping[str, Any]
    ) -> Mapping[str, torch.Tensor]:
        """Execute a training step given a minibatch of data.

        Set the model to train mode, run the forward pass, back propagate the
        gradients, step the optimizer, return the record for this step.

        Args:
            optimizer: optimizer of the model parameters.
            data_dict: a dict containing the current minibatch for training.
            hyper_param_dict: a dict containing hyperparameters for the current
                batch.
        
        Returns:
            A dict storing the record for this training step, which typically
            includes decomposed loss terms, gradient norm, and other values we
            care about.
        """

        self.train()
        optimizer.zero_grad()
        loss, _, new_record = self(data_dict, hyper_param_dict)
        loss.backward()
        norms = torch.nn.utils.clip_grad_norm_(self.parameters(), 500)
        new_record['max_norm'] = norms.cpu().numpy()
        optimizer.step()
        return new_record

    def get_embeddings_and_nll(self,
        adata: anndata.AnnData,
        batch_size: int = 2000,
        emb_names: Union[Iterable[str], None] = None,
        batch_col: str = 'batch_indices',
        inplace: bool = True
    ) -> Union[float, Tuple[Mapping[str, np.ndarray], float]]:
        """Calculate embeddings and nll for the given dataset.

        Args:
            adata: the test dataset. adata.n_vars must equal to #genes of this
                model.
            batch_size: batch size for test data input.
            emb_names: names of the embeddings to be returned or stored to
                adata.obsm.
            batch_col: a key in adata.obs to the batch column. Only used when
                self.need_batch is True.
            inplace: whether embeddings will be stored to adata or returned.

        Returns:
            If inplace, only the test nll. Otherwise, return the embeddings as
            a dict and also the test nll.
        """

        assert adata.n_vars == self.n_fixed_genes + self.n_trainable_genes
        assert not self.need_batch or adata.obs[batch_col].nunique() == self.n_batches
        if emb_names is None:
            emb_names = self.emb_names
        self.eval()
        if isinstance(emb_names, str):
            emb_names = [emb_names]

        sampler = CellSampler(adata, batch_size=batch_size, sample_batch_id=self.need_batch, n_epochs=1, batch_col=batch_col, shuffle=False)
        embs = {name: [] for name in emb_names}
        nll = 0.
        for data_dict in sampler:
            data_dict = {k: v.to(self.device) for k, v in data_dict.items()}
            fwd_dict = self(data_dict, dict(val=True))
            for name in emb_names:
                embs[name].append(fwd_dict[name].detach().cpu())
            nll += fwd_dict['nll'].detach().item()
        embs = {name: torch.cat(embs[name], dim=0).numpy() for name in emb_names}
        if inplace:
            for emb_name, emb in embs.items():
                adata.obsm[emb_name] = emb
        nll /= adata.n_obs

        if inplace:
            return nll
        else:
            return embs, nll



import os
import logging
import numpy as np
import scanpy as sc
import pandas as pd
import anndata as ad
import matplotlib.pyplot as plt
from scipy.sparse.csr import spmatrix
from scipy.stats import chi2
from typing import Sequence, Tuple, Iterable, Union
from scipy.sparse import issparse
from sklearn.metrics import adjusted_rand_score, normalized_mutual_info_score
from sklearn.neighbors import NearestNeighbors
import psutil

_cpu_count: Union[None, int] = psutil.cpu_count(logical=False)
if _cpu_count is None:
    _cpu_count: int = psutil.cpu_count(logical=True)
logger = logging.getLogger(__name__)


def _eff_n_jobs(n_jobs: int) -> int:
    """If n_jobs <= 0, set it as the number of physical cores _cpu_count"""
    return n_jobs if n_jobs > 0 else _cpu_count


def calculate_nearest_neighbors(
    X: np.array,
    n_neighbors: int = 100,
    n_jobs: int = -1,
    method: Union[str, None] = None,
    M: int = 20,
    efC: int = 200,
    efS: int = 200,
    random_state: int = 0,
    full_speed: int = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate nearest neighbors graph for the given data.

    Args:
        X: the sample-by-feature matrix.
        n_neighbors: # nearest neighbors to find.
        n_jobs: # jobs to generate. If <= 0, this is set to the number of
            physical cores.
        method: method to calculate the NN graph. Must be "hnsw" or "sklearn".
        M: maximum number of outgoing connections in the graph. Used only if
            method is "hnsw".
        efC: parameter that controls the index_time/index_accuracy. Bigger efC
            leads to longer construction, but better index quality. At some
            point, increasing efC does not improve the quality of the index.
            Used only if method is "hnsw".
        efS: the size of the dynamic list for the nearest neighbors (used
            during the search). Higher efS leads to more accurate but slower
            search. Can be anything between n_neighbors and #samples. Used only
            if method is "hnsw".
        random_state: random seed. Used only if method is "hnsw".
        full_speed: If full_speed, use multiple threads in constructing hnsw
            index. However, the kNN results are not reproducible. Used only if
            method is "hnsw".

    Returns:
        indicies: indices of the (n_neighbors - 1)-nearest neighbors, stored as
            a np.ndarray.
        distances: distances of the (n_neighbors - 1)-nearest neighbors, stored
            as a np.ndarray.
    """

    nsamples = X.shape[0]

    if method is None:
        if nsamples <= 1000:
            method = "sklearn"
        else:
            method = "hnsw"

    if nsamples < n_neighbors:
        logger.warning(f"Warning: in calculate_nearest_neighbors, number of samples = {nsamples} < n_neighbors = {n_neighbors}!\n Set n_neighbors to {nsamples}.")
        n_neighbors = nsamples

    n_jobs = _eff_n_jobs(n_jobs)

    if method == "hnsw":
        import hnswlib

        if issparse(X):
            X = X.toarray()
        # Build hnsw index
        knn_index = hnswlib.Index(space="l2", dim=X.shape[1])
        knn_index.init_index(
            max_elements=nsamples, ef_construction=efC, M=M, random_seed=random_state
        )
        knn_index.set_num_threads(n_jobs if full_speed else 1)
        knn_index.add_items(X)

        # KNN query
        knn_index.set_ef(efS)
        knn_index.set_num_threads(n_jobs)
        indices, distances = knn_index.knn_query(X, k=n_neighbors)
        # eliminate the first neighbor, which is the node itself
        if not (indices[:, 0] == np.arange(nsamples)).all():
            for i in range(nsamples):
                if indices[i, 0] != i:
                    indices[i, 1:] = indices[i, 0:-1]
                    distances[i, 1:] = distances[i, 0:-1]
        indices = indices[:, 1:].astype(int)
        distances = np.sqrt(distances[:, 1:])
    else:
        assert method == "sklearn"
        knn = NearestNeighbors(
            n_neighbors=n_neighbors - 1, n_jobs=n_jobs
        )  # eliminate the first neighbor, which is the node itself
        knn.fit(X)
        distances, indices = knn.kneighbors()

    return indices, distances


def _calculate_kbet_for_one_chunk(knn_indices, attr_values, ideal_dist, n_neighbors):
    dof = ideal_dist.size - 1

    ns = knn_indices.shape[0]
    results = np.zeros((ns, 2))
    for i in range(ns):
        observed_counts = (
            pd.Series(attr_values[knn_indices[i, :]]).value_counts(sort=False).values
        )
        expected_counts = ideal_dist * n_neighbors
        stat = np.sum(
            np.divide(
                np.square(np.subtract(observed_counts, expected_counts)),
                expected_counts,
            )
        )
        p_value = 1 - chi2.cdf(stat, dof)
        results[i, 0] = stat
        results[i, 1] = p_value

    return results


def _get_knn_indices(adata: ad.AnnData,
    use_rep: str = "delta",
    n_neighbors: int = 25,
    random_state: int = 0,
    full_speed: bool = False,
    n_jobs: int = -1,
    calc_knn: bool = True
) -> np.ndarray:

    nsample = adata.n_obs
    if calc_knn:
        assert use_rep in adata.obsm, f'{use_rep} not in adata.obsm'
        indices, distances = calculate_nearest_neighbors(
            adata.obsm[use_rep],
            n_neighbors=n_neighbors,
            n_jobs=_eff_n_jobs(n_jobs),
            random_state=random_state,
            full_speed=full_speed,
        )
    else:
        assert 'neighbors' in adata.uns, 'No precomputed knn exists.'
        assert adata.uns['neighbors']['params']['n_neighbors'] >= n_neighbors, f"pre-computed n_neighbors is {adata.uns['neighbors']['params']['n_neighbors']}, which is smaller than {n_neighbors}"
        indices = _knn_indices_from_dist(adata.obsp['distances'])
    knn_indices = np.concatenate(
        (np.arange(nsample).reshape(-1, 1), indices[:, 0 : n_neighbors - 1]), axis=1
    )  # add query as 1-nn
    return knn_indices


def calculate_kbet(
    adata: ad.AnnData,
    use_rep: str = "delta",
    batch_col: str = "batch_indices",
    n_neighbors: int = 25,
    alpha: float = 0.05,
    random_state: int = 0,
    full_speed: bool = False,
    n_jobs: int = -1,
    calc_knn: bool = True
) -> Tuple[float, float, float]:
    """Calculate the kBET metric of the data.

    kBET measures if cells from different batches mix well in their local
    neighborhood.

    Args:
        adata: annotated data matrix.
        use_rep: the embedding to be used. Must exist in adata.obsm.
        batch_col: a key in adata.obs to the batch column.
        n_neighbors: # nearest neighbors.
        alpha: acceptance rate threshold. A cell is accepted if its kBET
            p-value is greater than or equal to alpha.
        random_state: random seed. Used only if method is "hnsw".
        full_speed: If full_speed, use multiple threads in constructing hnsw
            index. However, the kNN results are not reproducible. Used only if
            method is "hnsw".
        n_jobs: # jobs to generate. If <= 0, this is set to the number of
            physical cores.
        calc_knn: whether to re-calculate the kNN graph or reuse the one stored
            in adata.

    Returns:
        stat_mean: mean kBET chi-square statistic over all cells.
        pvalue_mean: mean kBET p-value over all cells.
        accept_rate: kBET Acceptance rate of the sample.
    """

    logger.info('Calculating kbet...')
    assert batch_col in adata.obs
    assert adata.obs[batch_col].dtype.name == "category"

    ideal_dist = (
        adata.obs[batch_col].value_counts(normalize=True, sort=False).values
    )  # ideal no batch effect distribution
    nsample = adata.shape[0]
    nbatch = ideal_dist.size

    attr_values = adata.obs[batch_col].values.copy()
    attr_values.categories = range(nbatch)
    knn_indices = _get_knn_indices(adata, use_rep, n_neighbors, random_state, full_speed, n_jobs, calc_knn)

    # partition into chunks
    n_jobs = min(_eff_n_jobs(n_jobs), nsample)
    starts = np.zeros(n_jobs + 1, dtype=int)
    quotient = nsample // n_jobs
    remainder = nsample % n_jobs
    for i in range(n_jobs):
        starts[i + 1] = starts[i] + quotient + (1 if i < remainder else 0)

    from joblib import Parallel, delayed, parallel_backend
    with parallel_backend("loky", n_jobs=n_jobs):
        kBET_arr = np.concatenate(
            Parallel()(
                delayed(_calculate_kbet_for_one_chunk)(
                    knn_indices[starts[i] : starts[i + 1], :], attr_values, ideal_dist, n_neighbors
                )
                for i in range(n_jobs)
            )
        )

    res = kBET_arr.mean(axis=0)
    stat_mean = res[0]
    pvalue_mean = res[1]
    accept_rate = (kBET_arr[:, 1] >= alpha).sum() / nsample

    return (stat_mean, pvalue_mean, accept_rate)


def _entropy(hist_data):
    _, counts = np.unique(hist_data, return_counts = True)
    freqs = counts / counts.sum()
    return (-freqs * np.log(freqs + 1e-30)).sum()


def _entropy_batch_mixing_for_one_pool(batches, knn_indices, nsample, n_samples_per_pool):
    indices = np.random.choice(
        np.arange(nsample), size=n_samples_per_pool)
    return np.mean(
        [
            _entropy(batches[knn_indices[indices[i]]])
            for i in range(n_samples_per_pool)
        ]
    )


def calculate_entropy_batch_mixing(
    adata: ad.AnnData,
    use_rep: str = "delta",
    batch_col: str = "batch_indices",
    n_neighbors: int = 50,
    n_pools: int = 50,
    n_samples_per_pool: int = 100,
    random_state: int = 0,
    full_speed: bool = False,
    n_jobs: int = -1,
    calc_knn: bool = True
) -> float:
    """Calculate the entropy of batch mixing of the data.

    kBET measures if cells from different batches mix well in their local
    neighborhood.

    Args:
        adata: annotated data matrix.
        use_rep: the embedding to be used. Must exist in adata.obsm.
        batch_col: a key in adata.obs to the batch column.
        n_neighbors: # nearest neighbors.
        n_pools: #pools of cells to calculate entropy of batch mixing.
        n_samples_per_pool: #cells per pool to calculate within-pool entropy.
        random_state: random seed. Used only if method is "hnsw".
        full_speed: If full_speed, use multiple threads in constructing hnsw
            index. However, the kNN results are not reproducible. Used only if
            method is "hnsw".
        n_jobs: # jobs to generate. If <= 0, this is set to the number of
            physical cores.
        calc_knn: whether to re-calculate the kNN graph or reuse the one stored
            in adata.

    Returns:
        score: the mean entropy of batch mixing, averaged from n_pools samples.
    """

    logger.info('Calculating batch mixing entropy...')
    nsample = adata.n_obs

    knn_indices = _get_knn_indices(adata, use_rep, n_neighbors, random_state, full_speed, n_jobs, calc_knn)

    from joblib import Parallel, delayed, parallel_backend
    with parallel_backend("loky", inner_max_num_threads=1):
        score = np.mean(
            Parallel(n_jobs=n_jobs)(
                delayed(_entropy_batch_mixing_for_one_pool)(
                    adata.obs[batch_col], knn_indices, nsample, n_samples_per_pool
                )
                for _ in range(n_pools)
            )
        )
    return score


def clustering(
    adata: ad.AnnData,
    resolutions: Iterable[float],
    clustering_method: str = "leiden",
    cell_type_col: str = "cell_types",
    batch_col: str = "batch_indices"
) -> Tuple[str, float]:
    """Cluster the data and calculate agreement with cell type and batch
    variable.

    This method cluster the neighborhood graph (requires having run sc.pp.
    neighbors first) with "clustering_method" algorithm multiple times with the
    given resolutions, and return the best result in terms of ARI with cell
    type.
    Other metrics such as NMI with cell type, ARi with batch are logged but not
    returned. (TODO: also return these metrics)

    Args:
        adata: the dataset to be clustered. adata.obsp shouhld contain the keys
            'connectivities' and 'distances'.
        resolutions: a list of leiden/louvain resolution parameters. Will
            cluster with each resolution in the list and return the best result
            (in terms of ARI with cell type).
        clustering_method: Either "leiden" or "louvain".
        cell_type_col: a key in adata.obs to the cell type column.
        batch_col: a key in adata.obs to the batch column.

    Returns:
        best_cluster_key: a key in adata.obs to the best (in terms of ARI with
            cell type) cluster assignment column.
        best_ari: the best ARI with cell type.
        
    """

    if clustering_method == 'leiden':
        clustering_func: function = sc.tl.leiden
    elif clustering_method == 'louvain':
        clustering_func: function = sc.tl.louvain
    else:
        raise ValueError("Please specify louvain or leiden for the clustering method argument.")
    logger.info(f'Performing {clustering_method} clustering')
    assert cell_type_col in adata.obs, f"{cell_type_col} not in adata.obs"
    aris = []
    for res in resolutions:
        col = f'{clustering_method}_{res}'
        clustering_func(adata, resolution=res, key_added=col)
        ari = adjusted_rand_score(adata.obs[cell_type_col], adata.obs[col])
        nmi = normalized_mutual_info_score(adata.obs[cell_type_col], adata.obs[col])
        n_unique = adata.obs[col].nunique()
        aris.append((res, ari, n_unique))
        if batch_col in adata.obs and adata.obs[batch_col].nunique() > 1:
            ari_batch = adjusted_rand_score(adata.obs[batch_col], adata.obs[col])
            logger.info(f'Resolution: {res:5.3g}\tARI: {ari:7.4f}\tNMI: {nmi:7.4f}\tbARI: {ari_batch:7.4f}\t# labels: {n_unique}')
        else:
            logger.info(f'Resolution: {res:5.3g}\tARI: {ari:7.4f}\tNMI: {nmi:7.4f}\t# labels: {n_unique}')
    
    aris.sort(key=lambda x: x[1], reverse=True)
    best_res, best_ari = aris[0][0], aris[0][1]
    return f'{clustering_method}_{best_res}', best_ari


def draw_embeddings(adata: ad.AnnData,
        color_by: Union[str, Sequence[str], None] = None,
        min_dist: float = 0.3,
        spread: float = 1,
        save: bool = True,
        ckpt_dir: str = '.',
        fname: str = "umap.pdf",
        show: bool = False,
        dpi: int = 300,
    ) -> None:
    """Embed, plot and optionally save the neighborhood graph with UMAP.

    Requires having run sc.pp.neighbors first.

    Args:
        adata: the dataset to draw. adata.obsp shouhld contain the keys
            'connectivities' and 'distances'.
        color_by: a str or a list of adata.obs keys to color the points in the
            scatterplot by. E.g. if both cell_type_col and batch_col is in
            color_by, then we would have two plots colored by cell type and
            batch variables, respectively.
        min_dist: The effective minimum distance between embedded points.
            Smaller values will result in a more clustered/clumped embedding
            where nearby points on the manifold are drawn closer together,
            while larger values will result on a more even dispersal of points.
        spread: The effective scale of embedded points. In combination with
            `min_dist` this determines how clustered/clumped the embedded
            points are.
        save: whether to save the plot to ckpt_dir.
        ckpt_dir: where to save the plot. Only used if save = True.
        fname: file name of the saved plot. Only used if save = True.
        show: whether to show the plot (plt.show()). Useful for visualizing
            the plot in an interactive environment.
        dpi: the dpi of the saved plot. Only used if save = True.
    """

    sc.tl.umap(adata, min_dist=min_dist, spread=spread)
    fig = sc.pl.umap(adata, color=color_by, show=show, return_fig=True)
    if save:
        fig.savefig(
            os.path.join(ckpt_dir, fname),
            dpi=dpi, bbox_inches='tight'
        )
    fig.clf()
    plt.close(fig)


def _knn_indices_from_dist(dist: Union[spmatrix, np.ndarray]) -> Union[spmatrix, np.ndarray]:
    n_cells = dist.shape[0]
    return dist.nonzero()[1].reshape(n_cells, -1)



from typing import Sequence, Union
from torch import nn
import torch
import math

class InputPartlyTrainableLinear(nn.Module):
    """A linear layer with partially trainable input weights.

    The weights are divided into two parts, one of shape [I_trainable, O] is
    trainable, the other of shape [I_fixed, O] is fixed.
    If bias = True, the trainable part would have a trainbale bias, and if
    n_trainable_input is 0, a trainable bias is added to the layer.

    In the forward pass, the input x of shape [B, I] is split into x_fixed of
    shape [B, I_fixed] and x_trainable of shape [B, I_trainable]. The two parts
    are separately affinely transformed and results are summed.

    B: batch size; I: input dim; O: output dim.

    Attributes:
        fixed: the fixed part of the layer.
        trainable: the trainable part of the layer.
        trainable_bias: a trainable bias. Only present if n_trainable_input is
            0, i.e. all weights are fixed, and bias is True.
        n_fixed_input: number of inputs whose weights should be fixed.
        n_trainable_input: number of inputs whose weights should be trainable.
    """

    def __init__(self, n_fixed_input: int, n_output: int, n_trainable_input: int = 0, bias: bool = True) -> None:
        """Initialize the InputPartlyTrainableLinear layer.

        Args:
            n_fixed_input: number of inputs whose weights should be fixed.
            n_output: number of outputs.
            n_trainable_input: number of inputs whose weights should be
                trainable.
            bias: add a trainable bias if all weights are fixed. This gives
                more flexibility to the model.                
        """

        super().__init__()
        self.fixed: nn.Linear = nn.Linear(n_fixed_input, n_output, bias=False)
        self.fixed.requires_grad_(False)
        self.trainable_bias: Union[None, nn.Parameter] = None
        if n_trainable_input > 0:
            self.trainable: nn.Linear = nn.Linear(n_trainable_input, n_output, bias=bias)
        elif bias:
            self.trainable_bias = nn.Parameter(torch.Tensor(n_output))
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.fixed.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.trainable_bias, -bound, bound)
        self.n_fixed_input: int = n_fixed_input
        self.n_trainable_input: int = n_trainable_input

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the InputPartlyTrainableLinear layer.

        Args:
            x: the input tensor of shape [B, I].

        Returns:
            A linear-transformed x.
        """
        
        if self.n_trainable_input > 0:
            x_fixed, x_trainable = x[:, :self.n_fixed_input], x[:, self.n_fixed_input:]
            with torch.no_grad():
                out = self.fixed(x_fixed)
            return out + self.trainable(x_trainable)
        elif self.trainable_bias is not None:
            return self.fixed(x) + self.trainable_bias
        else:
            return self.fixed(x)

    @property
    def weight(self) -> torch.Tensor:
        """A read-only property to access the weights of this layer.

        If both trainable and fixed weights are present, concatenate them and
        return. Else return the fixed weights.
        """

        if self.n_trainable_input > 0:
            return torch.cat([self.fixed.weight, self.trainable.weight], dim=1)
        else:
            return self.fixed.weight
    
    @property
    def bias(self) -> Union[torch.Tensor, None]:
        """A read-only property to access the bias of this layer.

        If the trainable module exists, return the bias of the trainable
        module. Else, return self.trainable_bias (could be None).
        """

        if self.n_trainable_input > 0:
            return self.trainable.bias
        else:
            return self.trainable_bias

class OutputPartlyTrainableLinear(nn.Module):
    """A linear layer with partially trainable output weights.

    The weights are divided into two parts, one of shape [I, O_trainable] is
    trainable, the other of shape [I, O_fixed] is fixed.
    If bias = True, the trainable part would have a bias, and the fixed part
    would also have a trainble bias.

    In the forward pass, the input x of shape [B, I] is separately affinely
    transformed by the fixed and the trainable linear layers and the results
    are concatenated.

    B: batch size; I: input dim; O: output dim.

    Attributes:
        fixed: the fixed part of the layer.
        trainable: the trainable part of the layer.
        trainable_bias: a trainable bias. Only present if bias is True.
        n_fixed_output: number of outputs whose weights should be fixed.
        n_trainable_output: number of outputs whose weights should be
            trainable.
        enable_bias: whether the model has bias.
    """

    def __init__(self, n_input: int, n_fixed_output: int, n_trainable_output: int = 0, bias: bool = True) -> None:
        """Initialize the InputPartlyTrainableLinear layer.

        Args:
            n_fixed_input: number of inputs whose weights should be fixed.
            n_output: number of outputs.
            n_trainable_input: number of inputs whose weights should be
                trainable.
            bias: add a trainable bias if all weights are fixed. This gives
                more flexibility to the model.                
        """

        super().__init__(self)
        self.fixed: nn.Linear = nn.Linear(n_input, n_fixed_output, bias=False)
        self.fixed.requires_grad_(False)
        self.trainable_bias: Union[None, nn.Parameter] = None
        if bias:
            self.trainable_bias = nn.Parameter(torch.Tensor(n_fixed_output))
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.fixed.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.trainable_bias, -bound, bound)
        if n_trainable_output > 0:
            self.trainable: nn.Linear = nn.Linear(n_input, n_trainable_output, bias=bias)
        self.n_fixed_output: int = n_fixed_output
        self.n_trainable_output: int = n_trainable_output
        self.enable_bias: bool = bias

    def forward(self, x: torch.Tensor):
        """Forward pass of the OutputPartlyTrainableLinear layer.

        Args:
            x: the input tensor of shape [B, I].

        Returns:
            A linear-transformed x.
        """

        # calculate fixed output
        with torch.no_grad():
            fixed_output = self.fixed(x)
        if self.trainable_bias is not None:
            fixed_output = fixed_output + self.trainable_bias
        # return full output
        if self.n_trainable_output > 0:
            return torch.cat([fixed_output, self.trainable(x)], dim=-1) 
        else:
            return fixed_output

    @property
    def weight(self):
        """A read-only property to access the weights of this layer.

        If both trainable and fixed weights are present, concatenate them and
        return. Else return the fixed weights.
        """

        if self.n_trainable_output > 0:
            return torch.cat([self.fixed.weight, self.trainable.weight], dim=0)
        else:
            return self.fixed.weight

    @property
    def bias(self):
        """A read-only property to access the bias of this layer.

        If both trainable and fixed biases are present, concatenate them and
        return. Else, return fixed bias.
        """
        if not self.enable_bias:
            return None
        if self.n_trainable_output > 0:
            return torch.cat([self.trainable_bias, self.trainable.bias], dim=0)
        else:
            return self.trainable_bias

class PartlyTrainableParameter2D(nn.Module):
    """A partly trainable 2D parameter.

    The [H, W] parameter is split to two parts, the fixed [H, W_fixed] and the
    trainbale [H, W_trainable].

    H: height, W: width.

    Args:
        height: the height of the parameter.
        n_fixed_width: the width of the fixed part of the parameter.
        n_trainable_width: the width of the trainable part of the parameter.
        fixed: the fixed part of the parameter.
        trainable: the trainable part of the parameter.
    """

    def __init__(self, height: int, n_fixed_width: int, n_trainable_width: int) -> None:
        super().__init__()
        self.height: int = height
        self.n_fixed_width: int = n_fixed_width
        self.n_trainable_width: int = n_trainable_width
        self.fixed: Union[None, torch.Tensor] = None
        self.trainable: Union[None, nn.Parameter] = None
        if n_fixed_width > 0:
            self.fixed = torch.randn(height, n_fixed_width)
        if n_trainable_width > 0:
            self.trainable = nn.Parameter(torch.randn(height, n_trainable_width))
    
    def get_param(self) -> Union[None, torch.Tensor]:
        """Get the [H, W] parameter as a whole.

        Returns:
            The parameter. If both n_fixed_width and n_trainable_width is 0,
            return None.
        """

        params = [param for param in (self.fixed, self.trainable) if param is not None]
        if len(params) == 2:
            return torch.cat(params, dim=1)
        elif len(params) == 1:
            return params[0]
        else:
            return None

    def __repr__(self):
        return f'{self.__class__.__name__}(height={self.height}, fixed={self.n_fixed_width}, trainable={self.n_trainable_width})'


def get_fully_connected_layers(
    n_trainable_input: int,
    hidden_sizes: Union[int, Sequence[int]],
    n_trainable_output: Union[None, int] = None,
    bn: bool = True,
    bn_track_running_stats: bool = True,
    dropout_prob: float = 0.,
    n_fixed_input: int = 0,
    n_fixed_output: int = 0
) -> nn.Sequential:
    """Construct fully connected layers given the specifications.

    Args:
        n_trainable_input: number of trainable input.
        hidden_sizes: If int, constuct one hidden layer with the given
            size. If sequence of ints, construct a hidden layer for each
            element with the given size.
        n_trainable_output: number of trainable output. If None, do not
            construct the output layer.
        bn: whether to add a BatchNorm1d layer after ReLU activation.
        bn_track_running_stats: the track_running_stats argument of the
            nn.BatchNorm1d constructor.
        dropout_prob: dropout probability. If 0, disable dropout.
        n_fixed_input: number of fixed input. Parameters in the input layer
            related to these genes should be fixed. Useful for the fine-
            tuning stage in transfer learning.
        n_fixed_output: number of fixed output. Parameters in the output
            layer related to these genes should be fixed. Useful for the
            fine-tuning stage in transfer learning.
    
    Returns:
        The constructed fully connected layers.
    """

    if isinstance(hidden_sizes, int):
        hidden_sizes = [hidden_sizes]
    layers = []
    for i, size in enumerate(hidden_sizes):
        if i == 0 and n_fixed_input > 0:
            layers.append(InputPartlyTrainableLinear(n_fixed_input, size, n_trainable_input))
        else:
            layers.append(nn.Linear(n_trainable_input, size))
        layers.append(nn.ReLU())
        if bn:
            layers.append(nn.BatchNorm1d(size, track_running_stats=bn_track_running_stats))
        if dropout_prob:
            layers.append(nn.Dropout(dropout_prob))
        n_trainable_input = size
    if n_trainable_output is not None:
        if n_fixed_output > 0:
            layers.append(OutputPartlyTrainableLinear(n_trainable_input, n_fixed_output, n_trainable_output))
        else:
            layers.append(nn.Linear(n_trainable_input, n_trainable_output))
    return nn.Sequential(*layers)


def get_kl(mu: torch.Tensor, logsigma: torch.Tensor):
    """Calculate KL(q||p) where q = Normal(mu, sigma and p = Normal(0, I).

    Args:
        mu: the mean of the q distribution.
        logsigma: the log of the standard deviation of the q distribution.

    Returns:
        KL(q||p) where q = Normal(mu, sigma and p = Normal(0, I).
    """

    logsigma = 2 * logsigma
    return -0.5 * (1 + logsigma - mu.pow(2) - logsigma.exp()).sum(-1)